package search;

/*
 * This class performs the ant-based feature selection itself. A number of
 * ants are generated, which probabilistically traverse the graph of
 * features, storing their path (which corresponds to a reduct candidate).
 * An ant stops path construction when the gamma value (RS dependency metric)
 * is maximised or if the current path is too long. After this, the pheromone
 * is updated and the process continues again a fixed number of times
 *
 * (Note: any comment with RS by it indicates this is specific to rough sets)
 */
import java.util.BitSet;
import java.util.Iterator;
import java.util.Random;

import set.Subset;
import frfs2.*;

public final class AntSearch {

	Subset currReduct;
	int attributes; // number of attrs (includes the decision attr)
	int objects; // number of objects in the dataset
	BitSet remaining, empty, temp, t; // these are general use objects (RS)
	boolean verbose; // this is currently redundant
	Random random = new Random(); // random number generator

	double[][] table; // heuristic values that don't change

	// Ant-specific stuff
	int NO_OF_ANTS;// default set to = no of attrs
	double INITIAL_PHEROMONE = 0.5; // initial pheromone level for each edge
	int NO_OF_CYCLES = 270; // number of algorithmic iterations
	double MAX_HEURISTIC = 1;// need to calculate dataset gamma for this (RS)
	double MAX_GAMMA = 1; // same as above really (RS)
	int CUT_OFF = 30;// the most recent smallest reduct size

	double decay = 0.001;// the rate of pheromone decay
	Ant[] antz; // the array of ants
	double beta = 0.05;// beta value in the probabilistic transition rule
						// (heur)
	double alpha = 1; // alpha value in the probabilistic transition rule
						// (pher)
	double Q = 0.1; // used for pheromone update
	double q0 = 0.8; // used for a random deviation in an ant's path

	int heuristicType = -1; // which type of heuristic to use
	static final int NORMAL = 1;
	static final int ENTROPY = 2;
	static final int GAMMA = 3;

	int position = -1; // which method of initial ant placement to use
	static final int RANDOM = 1;
	static final int PER_FEATURE = 2;

	int updateType = -1; // which type of pheromone update
	static final int INDIVIDUAL_BEST = 1; // best ant updates it
	static final int ALL = 2; // all ants update pheromone
	static final int LOCAL = 3; // local update only (as in ACS)

	// pheromone table
	// e.g., to get the pheromone for edge (a,b) use pheromone[a][b]
	double[][] pheromone;
	Iterator<Edge> e;
	RSys2 rsys2;
	double quality;
	double c_divisor = 1;

	// constructor
	public AntSearch(RSys2 r, int o, int a, boolean v, int iterations,
			double al, double be, int ants, int ut, double q) {
		rsys2 = r;
		NO_OF_CYCLES = iterations;
		alpha = al;
		beta = be;

		objects = o;
		attributes = a;
		empty = new BitSet(objects);

		rsys2.setConsistency();

		// the matrix of pheromone levels
		pheromone = new double[a][a];

		// the number of ants to use...
		if (ants < 1)
			NO_OF_ANTS = attributes - 1;
		else
			NO_OF_ANTS = ants;

		antz = new Ant[NO_OF_ANTS];
		verbose = v;
		quality = q;

		// determines where the ants are placed
		// in this implementation, this is the only option
		position = PER_FEATURE;// RANDOM

		// intialise the ants
		init();
		setupHeuristics();

		// intialize all pheromone randomly
		randomPheromone();

		// determines how the pheromone is updated
		updateType = ut;

		MAX_HEURISTIC = 1; // max heuristic value for the dataset

		MAX_GAMMA = MAX_HEURISTIC;

		System.err.println("finished setup");
	}

	// construct array of heuristic values
	private final void setupHeuristics() {
		Subset bs = new Subset(attributes - 1);
		table = new double[attributes - 1][attributes - 1];

		// work out the heuristics...
		for (int a = 0; a < attributes - 1; a++) {
			table[a][a] = -1;
			// System.err.print(a+" ");

			for (int j = a + 1; j < attributes - 1; j++) {
				// the matrix is symmetric so if we've calculated
				// table[a][j] then we don't need to work out table[j][a]
				bs.add(a);
				bs.add(j);
				table[a][j] = rsys2.calculate(bs);
				table[j][a] = table[a][j];
				bs.remove(a);
				bs.remove(j);

			}

		}

	}

	/*
	 * Generate pheromone - varies randomly around the value 0.5 (the
	 * INITIAL_PHEROMONE level)
	 */
	public final void randomPheromone() {
		double max = 0;

		for (int i = 0; i < attributes - 1; i++) {
			pheromone[i][i] = 0.00001;

			for (int j = i + 1; j < attributes - 1; j++) {
				pheromone[i][j] = INITIAL_PHEROMONE + randomHeuristic() / 10;

				if (pheromone[i][j] == 0 || pheromone[i][j] >= 1)
					pheromone[i][j] = 0.1;
				if (pheromone[i][j] > max)
					max = pheromone[i][j];

				pheromone[j][i] = pheromone[i][j];
			}
		}
		normalise(max);
	}

	/*
	 * Initialise all ants
	 */
	private void init() {
		int attr = -1;
		switch (position) {

		// init ants - random (not used)
		case RANDOM:
			for (int i = 0; i < NO_OF_ANTS; i++) {
				attr = randomAttribute();
				antz[i] = new Ant(attr, i, attributes);
			}
			break;
		// init ants - per feature
		case PER_FEATURE:
			for (int a = 0; a < NO_OF_ANTS; a++) {
				// this is used where the number of ants>number of
				// conditional attributes
				if (a >= attributes - 1) {
					attr = randomAttribute();
					antz[a] = new Ant(attr, a, attributes);
				} else {
					antz[a] = new Ant(a, a, attributes);
				}
			}
			break;
		}
	}

	/*
	 * Heuristic evaluation of next attribute to be added. Could use gamma,
	 * entropy or some less costly evaluator.
	 */
	private double heuristic(Subset reduct, int current, int possible_next) {

		return table[current][possible_next];
	}

	// returns a random number between 0 and 1
	private double randomHeuristic() {
		return (double) (Math.abs(random.nextInt()) % (attributes))
				/ attributes;
	}

	// generate an initial random reduct for ants
	private final Subset randomReduct() {
		int attr;
		Subset ret = new Subset(attributes);

		// no of attrs appearing in initial reduct
		int no_of_attrs = 1;

		// To generate random reducts of size > 1 use the following code:
		// Math.abs(random.nextInt())%(attributes-1);
		// no_of_attrs = (int) no_of_attrs/3;
		// if (no_of_attrs==0) no_of_attrs=1;

		for (int i = 0; i < no_of_attrs; i++) {
			attr = Math.abs(random.nextInt()) % (attributes - 1);
			ret.add(attr);
		}

		return ret;
	}

	// returns a random conditional attribute number
	private final int randomAttribute() {
		return Math.abs(random.nextInt()) % (attributes - 1);
	}

	// decide how to update the pheromone
	private final void updatePheromone() {
		switch (updateType) {
		case INDIVIDUAL_BEST:
			individualBest();
			decay();
			break;
		case ALL:
			all();
			decay();
			break;
		case LOCAL:
			break;
		}

	}

	// this version finds the single best reduct and updates those edges
	private final void individualBest() {
		// for every ant, examine heuristic - best one updates pheromone
		int best_a = -1;
		double best_heur = -1;

		// find the single best ant
		for (int a = 0; a < NO_OF_ANTS; a++) {
			if (antz[a].heur > best_heur) { // this ends up being reduct's gamma
				best_heur = antz[a].heur;
				best_a = a;
			}
		}

		// get the set of edges of the best path
		e = antz[best_a].edges.iterator();
		int i, j;
		Edge ed = null;
		ed = (Edge) e.next();
		i = ed.i;

		// work out the size of this best reduct
		double s = antz[best_a].reduct.cardinality();
		if (s == 0)
			s = 99999;

		// update the pheromone along this path (traversing the edges)
		// according to the size of the reduct (smallest is best) and the
		// evaluation of the reduct (the gamma value stored in ant.heur)
		while (e.hasNext()) {
			ed = (Edge) e.next();
			j = ed.i;
			pheromone[i][j] += (Q * antz[best_a].heur * pheromone[i][j]) / s;

			// Math.min(1,pheromone[i][j]+= (Q*antz[best_a].heur)/s);
			// the pheromone matrix is symmetric
			pheromone[j][i] += (Q * antz[best_a].heur * pheromone[j][i]) / s;

			// Math.min(1,pheromone[j][i]+= (Q*antz[best_a].heur)/s)

			i = j;
		}
		normalise(-1.0);
	}

	/*
	 * Updates the pheromone using all ants Maybe alter this - at the minute it
	 * only looks at reducts whose gamma is maximum for the dataset. On each
	 * algorithmic iteration, it may be the case that *no* ant has found a
	 * reduct with gamma=1. It would be better to first cycle through the set of
	 * ants, find what their highest gamma value is, then get all reducts of
	 * this same gamma value and update the corresponding edges.
	 */
	private final void all() {
		for (int a = 0; a < NO_OF_ANTS; a++) {
			if (antz[a].heur == MAX_HEURISTIC) {
				e = antz[a].edges.iterator();
				int i, j;
				Edge ed = null;
				ed = (Edge) e.next();
				i = ed.i;
				double s = antz[a].reduct.cardinality();
				if (s == 0)
					s = 99999;

				while (e.hasNext()) {
					ed = (Edge) e.next();
					j = ed.i;
					// this assumes gamma=1 so ant.heur does not appear
					pheromone[i][j] += (pheromone[i][j]) / s;
					// Math.min(1,pheromone[i][j] + Q*(attributes-s)/a);
					pheromone[j][i] += (pheromone[j][i]) / s;
					// Math.min(1,pheromone[j][i] + Q*(attributes-s)/a);
					i = j;
				}
			}
		}
		normalise(-1.0);
	}

	public final void normalise(double m) {
		double max = m;

		if (max < 0) { // i.e. don't know the maximum
			for (int i = 0; i < attributes; i++) {
				for (int j = i; j < attributes; j++) {
					if (pheromone[i][j] > max)
						max = pheromone[i][j];
				}
			}
		}

		for (int i = 0; i < attributes; i++) {
			for (int j = i; j < attributes; j++) {
				pheromone[i][j] = pheromone[i][j] / max;
				pheromone[j][i] = pheromone[i][j]; // was /max
			}
		}
		// System.out.print(max+ " ");
	}

	// performs the pheromone decay
	public final void decay() {
		/*
		 * for (int a=0;a<attributes-1;a++) { for (int i=0;i<attributes-1;i++) {
		 * pheromone[a][i]*=(1-decay); } }
		 */
	}

	/*
	 * For the 'ant', construct the path using the probabilistic transition
	 * rule.
	 */
	private final void constructPath(Ant ant) {
		double best_prob = -1;
		int best_a = -1;
		double prob, heur;
		double gam = -1;
		boolean going = true;

		// while(gam<=MAX_HEURISTIC) {
		// if (!frfs2.reverse) {
		while (going) {
			best_a = -1;
			best_prob = -1;
			int prev = (ant.edges.get(ant.edges.size() - 1)).i;

			// random deviation - promotes extra exploration
			// this happens according to the value q0 - a high value of
			// q0 will limit random detours (a value of q0=1 will prevent
			// it altogether.
			if (Math.abs(random.nextDouble()) > q0) {
				while (true) {
					int poss = Math.abs(random.nextInt()) % (attributes - 1);
					best_a = poss;
					if (!ant.reduct.contains(best_a))
						break;
				}
				best_prob = 1;
			} else {
				// for each attribute, work out the value of the
				// probabilistic transition rule, choose the attribute
				// that produces its highest value
				for (int a = 0; a < attributes - 1; a++) {
					if (!ant.reduct.contains(a)) { // add an attribute
						Subset temp = (Subset) ant.reduct.clone();
						temp.add(a);

						// get the heuristic value
						heur = heuristic(temp, a, prev);
						if (heur == 0)
							heur = 0.000000001;

						// work out the probability using the heuristic
						// and pheromone levels
						prob = Math.pow(pheromone[a][prev], alpha)
								* Math.pow(heur, beta);

						// if this probability is better than the best so far
						// then consider this attribute to be the best
						if (prob > best_prob) {
							best_prob = prob;
							best_a = a;
						}

					}
				}
			}
			// if no best attribute has been found, then break
			if (best_a == -1) {
				break;
			}

			// set this attribute (it's the best one)
			ant.reduct.add(best_a);
			ant.addEdge(new Edge(best_a)); // remember path

			int reductsize = ant.reduct.cardinality();

			// evaluate the reduct
			gam = rsys2.calculate(ant.reduct);

			if (gam >= quality) {

				ant.heur = gam;
				if (reductsize < CUT_OFF) {
					CUT_OFF = reductsize;
					ant.reduct.add(attributes - 1);
					reductsize++;
					System.err.println(ant.reduct + " => " + gam + "  |"
							+ reductsize + "|");
					finalReduct = (Subset) ant.reduct.clone();
					ant.reduct.remove(attributes - 1);
					reductsize--;
				}
				currReduct = (Subset) ant.reduct.clone();
				ant.finished = true;
				going = false;
			} else if (gam < ant.heur || reductsize >= CUT_OFF) {
				ant.finished = true;
				going = false;
			}

			ant.heur = gam;
			if (updateType == LOCAL) {
				// pheromone[best_a][prev] =
				// (1-decay)*pheromone[best_a][prev]+(decay*INITIAL_PHEROMONE);
				pheromone[best_a][prev] = (1 - ant.heur) * INITIAL_PHEROMONE
						+ ant.heur * pheromone[best_a][prev];
				// pheromone[prev][best_a]= pheromone[best_a][prev];
			}
		}
	}

	Subset finalReduct = null;

	// main loop of the program for ant-based feature selection
	public final Subset run() {
		boolean going = true;
		int counter = 0;

		while (going) {
			counter++;

			// for each ant, if it hasn't finished then construct
			// a path
			for (int a = 0; a < NO_OF_ANTS; a++) {
				if (!antz[a].finished)
					constructPath(antz[a]);
			}

			// update the pheromone (and add decay)
			updatePheromone();

			// stop the process if the number of allowed
			// iterations is exceeded
			if (counter > NO_OF_CYCLES)
				going = false;

			// re-initialize the ants
			init();
			// if (verbose&&(counter%7==0)) printSolutions();
		}
		return finalReduct;
	}

	// this just prints pheromone for the time being
	private void printSolutions() {

		for (int a = 0; a < attributes - 1; a++) {
			System.err.print("[ ");
			for (int i = 0; i < attributes - 1; i++) {
				System.err.print(pheromone[a][i] + " ");
			}
			System.err.println("]");
		}
	}
}

class Distance implements Comparable<Object> {
	int object;
	double distance;
	int classIndex;

	public Distance(int o, double d, int c) {
		object = o;
		distance = d;
		classIndex = c;
	}

	// sort in ascending order (smallest first)
	public int compareTo(Object o) {
		Distance h = (Distance) o;
		return (int) (1000000 * (distance - h.distance));// (int)(gam-h.gam);
	}

}
